@@ -36,6 +36,7 @@
 #include "debug.h"
 
 #define READ_CHUNK_SIZE 8192
+#define MIME_READ_CHUNK_SIZE 1024
 
 struct GtkPlaylistPrivate
 {
@@ -74,6 +75,7 @@ enum {
 	PIX_COL,
 	FILENAME_COL,
 	URI_COL,
+	TITLE_CUSTOM_COL,
 	NUM_COLS
 };
 
@@ -159,6 +161,87 @@ gtk_tree_path_equals (GtkTreePath *path1, GtkTreePath *path2)
 	return retval;
 }
 
+/* This one returns a new string, in UTF8 even if the mrl is encoded
+ * in the locale's encoding
+ */
+static char *
+gtk_playlist_mrl_to_title (const gchar *mrl)
+{
+	char *filename_for_display, *filename, *unescaped;
+
+	filename = g_path_get_basename (mrl);
+	unescaped = gnome_vfs_unescape_string_for_display (filename);
+	g_free (filename);
+	filename_for_display = g_filename_to_utf8 (unescaped,
+			-1,             /* length */
+			NULL,           /* bytes_read */
+			NULL,           /* bytes_written */
+			NULL);          /* error */
+
+	g_free (unescaped);
+
+	return filename_for_display;
+}
+
+const char *
+my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data)
+{
+	GnomeVFSResult result;
+	GnomeVFSHandle *handle;
+	char *buffer;
+	const char *mimetype;
+	GnomeVFSFileSize total_bytes_read;
+	GnomeVFSFileSize bytes_read;
+
+	*data = NULL;
+
+	/* Open the file. */
+	result = gnome_vfs_open (&handle, uri, GNOME_VFS_OPEN_READ);
+	if (result != GNOME_VFS_OK) {
+		return NULL;
+	}
+
+	/* Read the whole thing. */
+	buffer = NULL;
+	total_bytes_read = 0;
+	do {
+		buffer = g_realloc (buffer, total_bytes_read
+				+ MIME_READ_CHUNK_SIZE);
+		result = gnome_vfs_read (handle,
+				buffer + total_bytes_read,
+				MIME_READ_CHUNK_SIZE,
+				&bytes_read);
+		if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF) {
+			g_free (buffer);
+			gnome_vfs_close (handle);
+			return NULL;
+		}
+
+		/* Check for overflow. */
+		if (total_bytes_read + bytes_read < total_bytes_read) {
+			g_free (buffer);
+			gnome_vfs_close (handle);
+			return NULL;
+		}
+
+		total_bytes_read += bytes_read;
+	} while (result == GNOME_VFS_OK
+			&& total_bytes_read < MIME_READ_CHUNK_SIZE);
+
+	/* Close the file. */
+	result = gnome_vfs_close (handle);
+	if (result != GNOME_VFS_OK) {
+		g_free (buffer);
+		return NULL;
+	}
+
+	/* Return the file. */
+	*data = g_realloc (buffer, total_bytes_read);
+	mimetype = gnome_vfs_get_mime_type_for_data (*data, total_bytes_read);
+
+	return mimetype;
+}
+
 static void
 gtk_tree_selection_has_selected_foreach (GtkTreeModel *model,
 		GtkTreePath *path, GtkTreeIter *iter, gpointer user_data)
@@ -182,7 +265,7 @@ gtk_tree_selection_has_selected (GtkTreeSelection *selection)
 }
 
 static GnomeVFSResult
-eel_read_entire_file (const char *uri,
+my_eel_read_entire_file (const char *uri,
 		int *file_size,
 		char **file_contents)
 {
@@ -236,6 +319,7 @@ eel_read_entire_file (const char *uri,
 	/* Return the file. */
 	*file_size = total_bytes_read;
 	*file_contents = g_realloc (buffer, total_bytes_read);
+
 	return GNOME_VFS_OK;
 }
 
@@ -373,15 +457,23 @@ static void
 selection_changed (GtkTreeSelection *treeselection, gpointer user_data)
 {
 	GtkPlaylist *playlist = GTK_PLAYLIST (user_data);
-	GtkWidget *remove_button;
+	GtkWidget *remove_button, *up_button, *down_button;
+	gboolean sensitivity;
 
 	remove_button = glade_xml_get_widget (playlist->_priv->xml,
 			"remove_button");
+	up_button = glade_xml_get_widget (playlist->_priv->xml, "up_button");
+	down_button = glade_xml_get_widget (playlist->_priv->xml,
+			"down_button");
 
 	if (gtk_tree_selection_has_selected (treeselection))
-		gtk_widget_set_sensitive (remove_button, TRUE);
+		sensitivity = TRUE;
 	else
-		gtk_widget_set_sensitive (remove_button, FALSE);
+		sensitivity = FALSE;
+
+	gtk_widget_set_sensitive (remove_button, sensitivity);
+	gtk_widget_set_sensitive (up_button, sensitivity);
+	gtk_widget_set_sensitive (down_button, sensitivity);
 }
 
 /* This function checks if the current item is NULL, and try to update it as the
@@ -588,7 +680,8 @@ init_treeview (GtkWidget *treeview, GtkPlaylist *playlist)
 	model = GTK_TREE_MODEL (gtk_list_store_new (NUM_COLS,
 				GDK_TYPE_PIXBUF,
 				G_TYPE_STRING,
-				G_TYPE_STRING));
+				G_TYPE_STRING,
+				G_TYPE_BOOLEAN));
 
 	/* the treeview */
 	gtk_tree_view_set_model (GTK_TREE_VIEW (treeview), model);
@@ -854,17 +947,7 @@ gtk_playlist_add_one_mrl (GtkPlaylist *playlist, const char *mrl,
 
 	if (display_name == NULL)
 	{
-		char *filename, *unescaped;
-
-		filename = g_path_get_basename (mrl);
-		unescaped = gnome_vfs_unescape_string_for_display (filename);
-		g_free (filename);
-		filename_for_display = g_filename_to_utf8 (unescaped,
-				-1,		/* length */
-				NULL,		/* bytes_read */
-				NULL,		/* bytes_written */
-				NULL);		/* error */
-		g_free (unescaped);
+		filename_for_display = gtk_playlist_mrl_to_title (mrl);
 	} else {
 		filename_for_display = g_strdup (display_name);
 	}
@@ -875,6 +958,7 @@ gtk_playlist_add_one_mrl (GtkPlaylist *playlist, const char *mrl,
 			PIX_COL, NULL,
 			FILENAME_COL, filename_for_display,
 			URI_COL, mrl,
+			TITLE_CUSTOM_COL, display_name ? TRUE : FALSE,
 			-1);
 
 	g_free (filename_for_display);
@@ -897,15 +981,11 @@ gtk_playlist_add_m3u (GtkPlaylist *playlist, const char *mrl)
 	char *contents, **lines;
 	int size, i;
 
-	if (eel_read_entire_file (mrl, &size, &contents) != GNOME_VFS_OK)
+	if (my_eel_read_entire_file (mrl, &size, &contents) != GNOME_VFS_OK)
 		return FALSE;
 
-	if (strstr (contents, "\n") == NULL)
-	{
-		retval = gtk_playlist_add_one_mrl (playlist, contents, NULL);
-		g_free (contents);
-		return retval;
-	}
+	contents = g_realloc (contents, size + 1);
+	contents[size] = '\0';
 
 	lines = g_strsplit (contents, "\n", 0);
 	g_free (contents);
@@ -916,7 +996,9 @@ gtk_playlist_add_m3u (GtkPlaylist *playlist, const char *mrl)
 		if (strstr(lines[i], "://") != NULL
 				|| lines[i][0] == G_DIR_SEPARATOR)
 		{
-			if (gtk_playlist_add_one_mrl (playlist,
+			/* We use the same code for .ram and m3u playlists,
+			 * and .ram files can contain .smil entries */
+			if (gtk_playlist_add_mrl (playlist,
 						lines[i], NULL) == TRUE)
 				retval = TRUE;
 		}
@@ -927,6 +1009,23 @@ gtk_playlist_add_m3u (GtkPlaylist *playlist, const char *mrl)
 	return retval;
 }
 
+static gboolean
+gtk_playlist_add_asf_playlist (GtkPlaylist *playlist, const char *mrl)
+{
+	gboolean retval = FALSE;
+	char *contents, **lines;
+#if 0
+	if (my_eel_read_entire_file (mrl, &size, &contents) != GNOME_VFS_OK)
+		return FALSE;
+
+	contents = g_realloc (contents, size + 1);
+	contents[size] = '\0';
+
+	lines = g_strsplit (contents, "\n", 0);
+	g_free (contents);
+#endif
+}
+
 static gboolean
 gtk_playlist_add_pls (GtkPlaylist *playlist, const char *mrl)
 {
@@ -934,9 +1033,12 @@ gtk_playlist_add_pls (GtkPlaylist *playlist, const char *mrl)
 	char *contents, **lines;
 	int size, i, num_entries;
 
-	if (eel_read_entire_file (mrl, &size, &contents) != GNOME_VFS_OK)
+	if (my_eel_read_entire_file (mrl, &size, &contents) != GNOME_VFS_OK)
 		return FALSE;
 
+	contents = g_realloc (contents, size + 1);
+	contents[size] = '\0';
+
 	lines = g_strsplit (contents, "\n", 0);
 	g_free (contents);
 
@@ -983,7 +1085,7 @@ bail:
 }
 
 static gboolean
-parse_entry (GtkPlaylist *playlist, char *base, xmlDocPtr doc,
+parse_asx_entry (GtkPlaylist *playlist, char *base, xmlDocPtr doc,
 		xmlNodePtr parent)
 {
 	xmlNodePtr node;
@@ -1034,7 +1136,7 @@ parse_entry (GtkPlaylist *playlist, char *base, xmlDocPtr doc,
 }
 
 static gboolean
-parse_entries (GtkPlaylist *playlist, char *base, xmlDocPtr doc,
+parse_asx_entries (GtkPlaylist *playlist, char *base, xmlDocPtr doc,
 		xmlNodePtr parent)
 {
 	xmlNodePtr node;
@@ -1048,7 +1150,7 @@ parse_entries (GtkPlaylist *playlist, char *base, xmlDocPtr doc,
 		if (g_ascii_strcasecmp (node->name, "entry") == 0)
 		{
 			/* Whee found an entry here, find the REF and TITLE */
-			if (parse_entry (playlist, base, doc, node) == TRUE)
+			if (parse_asx_entry (playlist, base, doc, node) == TRUE)
 				retval = TRUE;
 		}
 	}
@@ -1065,16 +1167,22 @@ gtk_playlist_add_asx (GtkPlaylist *playlist, const char *mrl)
 	int size;
 	gboolean retval = FALSE;
 
-	if (eel_read_entire_file (mrl, &size, &contents) != GNOME_VFS_OK)
+	if (my_eel_read_entire_file (mrl, &size, &contents) != GNOME_VFS_OK)
 		return FALSE;
 
-	doc = xmlParseMemory(contents, size);
+	contents = g_realloc (contents, size + 1);
+	contents[size] = '\0';
+
+	doc = xmlParseMemory (contents, size);
+	if (doc == NULL)
+		doc = xmlRecoverMemory (contents, size);
 	g_free (contents);
 
 	/* If the document has no root, or no name */
-	if(!doc->children || !doc->children->name)
+	if(!doc || !doc->children || !doc->children->name)
 	{
-		xmlFreeDoc(doc);
+		if (doc != NULL)
+			xmlFreeDoc(doc);
 		return FALSE;
 	}
 
@@ -1089,9 +1197,9 @@ gtk_playlist_add_asx (GtkPlaylist *playlist, const char *mrl)
 		gnome_vfs_uri_unref (parent);
 	}
 
-	for(node = doc->children; node != NULL; node = node->next)
+	for (node = doc->children; node != NULL; node = node->next)
 	{
-		if (parse_entries (playlist, base, doc, node) == TRUE)
+		if (parse_asx_entries (playlist, base, doc, node) == TRUE)
 			retval = TRUE;
 	}
 
@@ -1101,55 +1209,233 @@ gtk_playlist_add_asx (GtkPlaylist *playlist, const char *mrl)
 }
 
 static gboolean
-gtk_playlist_add_ra (GtkPlaylist *playlist, const char *mrl)
+gtk_playlist_add_ra (GtkPlaylist *playlist, const char *mrl, gpointer data)
 {
+	if (data == NULL
+			|| (strncmp (data, "http://", strlen ("http://")) != 0
+			&& strncmp (data, "rtsp://", strlen ("rtsp://")) != 0
+			&& strncmp (data, "pnm://", strlen ("pnm://")) != 0))
+	{
+		return gtk_playlist_add_one_mrl (playlist, mrl, NULL);
+	}
+
 	/* How nice, same format as m3u it seems */
 	return gtk_playlist_add_m3u (playlist, mrl);
 }
-#if 0 //FIXME
+
 static gboolean
-gtk_playlist_add_wmv (GtkPlaylist *playlist, const char *mrl)
+parse_smil_video_entry (GtkPlaylist *playlist, char *base,
+		char *url, char *title)
 {
+	gboolean retval = FALSE;
+
+	if (strstr (url, "://") != NULL || url[0] == '/')
+	{
+		retval = gtk_playlist_add_one_mrl (playlist, url, title);
+	} else {
+		char *fullpath;
+
+		fullpath = g_strdup_printf ("%s/%s", base, url);
+		retval = gtk_playlist_add_one_mrl (playlist, fullpath, title);
+
+		g_free (fullpath);
+	}
+
+	return retval;
 }
+
+static gboolean
+parse_smil_entry (GtkPlaylist *playlist, char *base, xmlDocPtr doc,
+		xmlNodePtr parent)
+{
+	xmlNodePtr node;
+	char *title, *url;
+	gboolean retval = FALSE;
+
+	title = NULL;
+	url = NULL;
+
+	for (node = parent->children; node != NULL; node = node->next)
+	{
+		if (node->name == NULL)
+			continue;
+
+		/* ENTRY should only have one ref and one title nodes */
+		if (g_ascii_strcasecmp (node->name, "video") == 0)
+		{
+			url = xmlGetProp (node, "src");
+			title = xmlGetProp (node, "title");
+
+			if (url != NULL)
+			{
+				if (parse_smil_video_entry (playlist,
+						base, url, title) == TRUE)
+					retval = TRUE;
+			}
+
+			g_free (title);
+			g_free (url);
+		} else {
+			if (parse_smil_entry (playlist,
+						base, doc, node) == TRUE)
+			{
+				retval = TRUE;
+			}
+		}
+	}
+
+	return retval;
+}
+
+static gboolean
+parse_smil_entries (GtkPlaylist *playlist, char *base, xmlDocPtr doc,
+		xmlNodePtr parent)
+{
+	xmlNodePtr node;
+	gboolean retval = FALSE;
+
+	for (node = parent->children; node != NULL; node = node->next)
+	{
+		if (node->name == NULL)
+			continue;
+
+		if (g_ascii_strcasecmp (node->name, "body") == 0)
+		{
+			if (parse_smil_entry (playlist, base,
+						doc, node) == TRUE)
+			{
+				retval = TRUE;
+			}
+		}
+
+	}
+
+	return retval;
+}
+
+static gboolean
+gtk_playlist_add_smil (GtkPlaylist *playlist, const char *mrl)
+{
+	xmlDocPtr doc;
+	xmlNodePtr node;
+	char *contents = NULL, *base;
+	int size;
+	gboolean retval = FALSE;
+
+	if (my_eel_read_entire_file (mrl, &size, &contents) != GNOME_VFS_OK)
+		return FALSE;
+
+	contents = g_realloc (contents, size + 1);
+	contents[size] = '\0';
+
+	doc = xmlParseMemory (contents, size);
+	if (doc == NULL)
+		doc = xmlRecoverMemory (contents, size);
+	g_free (contents);
+
+	/* If the document has no root, or no name */
+	if(!doc || !doc->children
+			|| !doc->children->name
+			|| g_ascii_strcasecmp (doc->children->name,
+				"smil") != 0)
+	{
+		if (doc != NULL)
+			xmlFreeDoc(doc);
+		return FALSE;
+	}
+
+	/* Yay, let's reconstruct the base by hand */
+	{
+		GnomeVFSURI *uri, *parent;
+		uri = gnome_vfs_uri_new (mrl);
+		parent = gnome_vfs_uri_get_parent (uri);
+		base = gnome_vfs_uri_to_string (parent, 0);
+
+		gnome_vfs_uri_unref (uri);
+		gnome_vfs_uri_unref (parent);
+	}
+
+	for (node = doc->children; node != NULL; node = node->next)
+	{
+		if (parse_smil_entries (playlist, base, doc, node) == TRUE)
+			retval = TRUE;
+	}
+
+	return FALSE;
+}
+
+static gboolean
+gtk_playlist_add_asf (GtkPlaylist *playlist, const char *mrl, gpointer data)
+{
+#if 0
+	if (data == NULL)
+		return gtk_playlist_add_one_mrl (playlist, mrl);
+
+	if (strncmp (data, "[Reference]", strlen ("[Reference]")) != 0)
+			return gtk_playlist_add_one_mrl (playlist, mrl);
+
+	return gtk_playlist_add_asf_playlist (playlist, mrl);
 #endif
+}
+
 gboolean
 gtk_playlist_add_mrl (GtkPlaylist *playlist, const char *mrl,
 		const char *display_name)
 {
 	const char *mimetype;
+	gpointer data;
+	gboolean retval = FALSE;
 
 	g_return_val_if_fail (mrl != NULL, FALSE);
 
-	mimetype = gnome_vfs_get_mime_type (mrl);
+	mimetype = my_gnome_vfs_get_mime_type_with_data (mrl, &data);
 
 	if (mimetype == NULL)
 	{
-		D("trying to add '%s' with no mimetype", mrl);
+		D("adding '%s' with no mimetype", mrl);
+		g_free (data);
 		return gtk_playlist_add_one_mrl (playlist, mrl, display_name);
 	} else if (strcmp ("audio/x-mpegurl", mimetype) == 0) {
+		g_free (data);
 		return gtk_playlist_add_m3u (playlist, mrl);
 	} else if (strcmp ("audio/x-scpls", mimetype) == 0) {
+		g_free (data);
 		return gtk_playlist_add_pls (playlist, mrl);
 	} else if (strcmp ("audio/x-ms-asx", mimetype) == 0) {
 		return gtk_playlist_add_asx (playlist, mrl);
 	} else if (strcmp ("audio/x-real-audio", mimetype) == 0
-			|| strcmp ("audio/x-pn-realaudio", mimetype) == 0) {
-		return gtk_playlist_add_ra (playlist, mrl);
+			|| strcmp ("audio/x-pn-realaudio", mimetype) == 0
+			|| strcmp ("application/vnd.rn-realmedia",
+				mimetype) == 0
+			|| strcmp ("audio/x-pn-realaudio-plugin",
+				mimetype) == 0) {
+		retval = gtk_playlist_add_ra (playlist, mrl, data);
+		g_free (data);
+		return retval;
+	} else if (strcmp ("video/x-ms-asf", mimetype) == 0) {
+		retval = gtk_playlist_add_asf (playlist, mrl, data);
+		g_free (data);
+		return retval;
+	} else if (strcmp ("application/x-smil", mimetype) == 0) {
+		g_free (data);
+		return gtk_playlist_add_smil (playlist, mrl);
 	} else if (strcmp ("x-directory/normal", mimetype) == 0) {
 		//FIXME Load all the files in the dir ?
 	}
 
+	g_free (data);
+
 	if (strncmp ("audio/", mimetype, 6) != 0
 			&& strncmp ("video/", mimetype, 6) != 0
-			&& strncmp ("application/x-ogg", mimetype, 17) != 0)
+			&& strncmp ("application/x-ogg", mimetype, 17) != 0
+			&& strncmp ("image/png", mimetype, 9) != 0)
 	{
-		D("trying to add '%s' with mimetype '%s'",
+		//FIXME error message
+		D("not adding '%s' with mimetype '%s'",
 				mrl, mimetype);
 		return FALSE;
 	}
 
-	//FIXME check size != 0
-
 	return gtk_playlist_add_one_mrl (playlist, mrl, display_name);
 }
 
@@ -1190,6 +1476,30 @@ char
 	return path;
 }
 
+char
+*gtk_playlist_get_current_title (GtkPlaylist *playlist, gboolean *custom)
+{
+	GtkTreeIter iter;
+	char *path;
+
+	g_return_val_if_fail (GTK_IS_PLAYLIST (playlist), NULL);
+
+	if (update_current_from_playlist (playlist) == FALSE)
+		return NULL;
+
+	gtk_tree_model_get_iter (playlist->_priv->model,
+			&iter,
+			playlist->_priv->current);
+
+	gtk_tree_model_get (playlist->_priv->model,
+			&iter,
+			FILENAME_COL, &path,
+			TITLE_CUSTOM_COL, custom,
+			-1);
+
+	return path;
+}
+
 gboolean
 gtk_playlist_has_previous_mrl (GtkPlaylist *playlist)
 {
@@ -1422,40 +1732,6 @@ gtk_playlist_set_at_end (GtkPlaylist *playlist)
 	}
 }
 
-/* This one returns a new string, in UTF8 even if the mrl is encoded
- * in the locale's encoding
- */
-gchar *
-gtk_playlist_mrl_to_title (const gchar *mrl)
-{
-	char *filename_for_display, *with_suffix, *filename, *unescaped;
-
-	filename = g_path_get_basename (mrl);
-	unescaped = gnome_vfs_unescape_string_for_display (filename);
-	g_free (filename);
-	with_suffix = g_filename_to_utf8 (unescaped,
-			-1,             /* length */
-			NULL,           /* bytes_read */
-			NULL,           /* bytes_written */
-			NULL);          /* error */
-
-	g_free (unescaped);
-
-	if (strrchr (with_suffix, '.')
-			&& strlen (strrchr (with_suffix, '.')) < 5)
-	{
-		filename_for_display = g_strndup (with_suffix,
-				strlen (with_suffix)
-				- strlen (strrchr (with_suffix, '.')));
-	} else {
-		filename_for_display = g_strdup (with_suffix);
-	}
-
-	g_free (with_suffix);
-
-	return filename_for_display;
-}
-
 static void
 gtk_playlist_class_init (GtkPlaylistClass *klass)
 {
