@@ -21,7 +21,8 @@
 #define __TOTEM_H__
 
 #include <gconf/gconf-client.h>
-#include "gtk-xine.h"
+#include <gtk/gtkwindow.h>
+#include "bacon-video-widget.h"
 
 #define TOTEM_GCONF_PREFIX "/apps/totem"
 
@@ -37,14 +38,15 @@ void	totem_action_next			(Totem *totem);
 void	totem_action_previous			(Totem *totem);
 void	totem_action_seek_relative		(Totem *totem, int off_sec);
 void	totem_action_volume_relative		(Totem *totem, int off_pct);
-void	totem_action_open_files			(Totem *totem, char **list,
+gboolean totem_action_open_files		(Totem *totem, char **list,
 						 gboolean ignore_first);
-void	totem_action_set_mrl			(Totem *totem, const char *mrl);
+gboolean totem_action_set_mrl			(Totem *totem, const char *mrl);
 
 void	totem_action_play_media			(Totem *totem, MediaType type);
 
 void	totem_action_toggle_aspect_ratio	(Totem *totem);
 void	totem_action_set_scale_ratio		(Totem *totem, gfloat ratio);
+void    totem_action_error                      (char *msg, GtkWindow *parent);
 
 GConfClient *totem_get_gconf_client		(Totem *totem);
 
